;; Generated package description from org-remark.el  -*- no-byte-compile: t -*-
(define-package "org-remark" "0.2.0.0.20220205.145952" "Highlight & annotate any text files" '((emacs "27.1") (org "9.4")) :commit "f713a3846ba11c7e7939f8f2aeb229efae8401f2" :authors '(("Noboru Ota" . "me@nobiot.com")) :maintainer '("Noboru Ota" . "me@nobiot.com") :keywords '("org-mode" "annotation" "writing" "note-taking" "marginal-notes") :url "https://github.com/nobiot/org-remark")
